/********************************************************* begin of preamble
**
** Copyright (C) 2003-2012 Software- und Organisations-Service GmbH. 
** All rights reserved.
**
** This file may be used under the terms of either the 
**
**   GNU General Public License version 2.0 (GPL)
**
**   as published by the Free Software Foundation
**   http://www.gnu.org/licenses/gpl-2.0.txt and appearing in the file
**   LICENSE.GPL included in the packaging of this file. 
**
** or the
**  
**   Agreement for Purchase and Licensing
**
**   as offered by Software- und Organisations-Service GmbH
**   in the respective terms of supply that ship with this file.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
** IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
** THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
** BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
** CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
********************************************************** end of preamble*/
#ifndef __SCHEDULER_TIMED_CALL_H
#define __SCHEDULER_TIMED_CALL_H

#include "../zschimmer/Call.h"

namespace sos {
namespace scheduler {

//---------------------------------------------------------------------------------------Timed_call
// Ein Aufruf zu einer bestimmten Zeit

struct Timed_call : z::Call, javabridge::has_proxy<Timed_call> 
{
  private:
    Time _at;

  protected:
    Timed_call(const Time& at) : _at(at) {}

  public:
    int64 at_millis() const { return at().millis(); }
    
    Time at() const { return _at; }

    virtual void call() const = 0;
};

//--------------------------------------------------------------------------------------Object_call
// Ein Aufruf eines Scheduler_object zu einer bestimmten Zeit

struct Object_call : Timed_call {
  protected: 
    Object_call(const Time& at = Time(0)) : Timed_call(at) {}

    string obj_name() const;

  protected:
    virtual Scheduler_object* object() const = 0;

    virtual string call_name() const { return name_of_type_info(typeid(*this)); }
};

//-------------------------------------------------------------------------------------Type_int_map

struct Type_int_map {
    static Type_int_map static_singleton;

  private:
    Mutex _mutex;
    stdext::hash_map<string, int> _map;

  public:
    int type_to_int(const type_info&);
};

//------------------------------------------------------------------------------------object_call<>
// Ein Aufruf einer Methode OBJECT::on_call(const CALL&) zu einer bestimmten Zeit

template<typename OBJECT, typename CALL>
struct object_call : Object_call {
  private: 
    OBJECT* const _object;

  protected:
    object_call(OBJECT* o) : _object(o) {}
    object_call(const Time& at, OBJECT* o) : Object_call(at), _object(o) {}

    Scheduler_object* object() const { return _object; }

  public:
    void call() const {
        _object->on_call(*(const CALL*)this);
    }

    static int type_id() { return Type_int_map::static_singleton.type_to_int(typeid(CALL)); }
};

//-------------------------------------------------------------------------------DEFINE_SIMPLE_CALL
// Definiert einen Klasse fr den parameterlosen Aufruf

#define DEFINE_SIMPLE_CALL(OBJECT_TYPE, CALL) \
    struct CALL : object_call<OBJECT_TYPE, CALL> { \
        CALL(OBJECT_TYPE* o) : object_call<OBJECT_TYPE, CALL>(Time(0), o) {} \
        CALL(const Time& at, OBJECT_TYPE* o) : object_call<OBJECT_TYPE, CALL>(at, o) {} \
    };


//struct Operation_holder {
//  private:
//    Spooler* _spooler;
//    ptr<Timed_call> _operation;
//
//  public:
//    Operation_holder(Spooler*);
//
//    void enqueue(Timed_call*);
//    void cancel();
//    Time at() const;
//};
//
//
//
//template<typename OBJECT, typename OPERATION>
//struct simple_operation_holder {
//private:
//    OBJECT* _object;
//    Operation_holder _holder;
//
//  public:
//    simple_operation_holder(OBJECT* o) : _object(o), _holder(o->spooler()) {}
//
//    ~simple_operation_holder() {
//        cancel();
//    }
//
//    void call_at(const Time& at) {
//        if (at.is_never())
//            _holder.cancel();
//        else
//            _holder.enqueue(Z_NEW(OPERATION(at, _object)));
//    }
//
//    void cancel() {
//        _holder.cancel();
//    }
//
//    //Time at() const {
//    //    return _holder.at();
//    //}
//};

//------------------------------------------------------------------------------Typed_call_register

struct Typed_call_register {
  private:
    Spooler* _spooler;
    typedef stdext::hash_map<int, ptr<Timed_call> > Map;
    Map _map;

  protected:
    Typed_call_register(Spooler*);

  public:
    ~Typed_call_register();

    Time next_time() const;

  protected:
    void enqueue_id(int id, Timed_call* o);
    void cancel_id(int id);
    void cancel_entry(ptr<Timed_call>*);
};

//----------------------------------------------------------------------------typed_call_register<>

template<typename OBJECT>
struct typed_call_register : Typed_call_register {
  private:
    OBJECT* _object;
    
  public:
    typed_call_register(OBJECT* o) : Typed_call_register(o->spooler()), _object(o) {}

    template<typename CALL>
    void cancel() {
        cancel_id(CALL::type_id());
    }
    
    template<typename CALL>
    void call_at(const Time& at) {
        if (at.is_never()) cancel<CALL>();
        else call<CALL>(Z_NEW(CALL(at, _object)));
    }
    
    template<typename CALL>
    void call() {
        call<CALL>(Z_NEW(CALL(Time(0), _object)));
    }
    
    template<typename CALL>
    void call(CALL* call) {
        enqueue_id(CALL::type_id(), call);
    }

    template<typename CALL>
    void call(const ptr<CALL>& call) {
        enqueue_id(CALL::type_id(), call);
    }
};

}}

#endif
